/**
 * diam_tcp_mngmt.js
 * This file is part of the YATE Project http://YATE.null.ro
 *
 * Yet Another Telephony Engine - a fully featured software PBX and IVR
 * Copyright (C) 2016 Null Team
 *
 * This software is distributed under multiple licenses;
 * see the COPYING file in the main directory for licensing
 * information for this specific distribution.
 *
 * This use of this software may be subject to additional restrictions.
 * See the LEGAL file in the main directory for details.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * This script handles managament messages regarding Diameter over TCP connections
 *
 * [scripts]
 * diam_tcp_mngmt=diam_tcp_mngmt.js
 */

function DiameterConn(connName)
{
    this.name = connName;
    this.state = "init";
    this.capabs = "";
}

DiameterConn.prototype = new Object;

DiameterConn.capabs = {
    HSSMobility : { appId : "16777251",
		    operations: "UpdateLocationRequest,ULR,UpdateLocationAnswer,ULA,PurgeUERequest,PUR,PurgeUEAnswer,PUA,NotifyRequest,NOR,NotifyAnswer,NOA"},
    MMEMobility : { appId : "16777251",
		    operations: "CancelLocationRequest,CLR,CancelLocationAnswer,CLA,InsertSubscriberDataRequest,IDR,InsertSubscriberDataAnswer,IDA,DeleteSubscriberDataRequest,DSR,DeleteSubscriberDataAnswer,DSA,ResetRequest,RSR,ResetAnswer,RSA"},
    HSSAuthentication : { appId : "16777251",
		    operations: "AuthenticationInfoRequest,AIR,AuthenticationInfoAnswer,AIA"},
    None : { appId : "", operations : "" },
};

DiameterConn.xmlns = "http://yate.null.ro/xml/diameter/v1";
DiameterConn.xmlDeclEncoding = "UTF-8";
DiameterConn.xmlDeclVersion = "1.0";
DiameterConn.xmlAsText = true;
DiameterConn.connList = {};


DiameterConn.prototype.buildMsg = function(name,xml)
{
    var msg = new Message(name);
    msg.module = "diam_tcp";
    msg.application = this.name;
    if (xml) {
	if (xml.getTag)
	    xml.put(msg,"xml",DiameterConn.xmlAsText);
    }
    return msg;
};

DiameterConn.prototype.handleXml = function(xml,msg,checkXmlns)
{
    if (!xml)
	return false;
    if (!xml.getTag)
	return false;
    if (checkXmlns) {
	var xmlns = xml.getAttribute("xmlns");
	if (xmlns != DiameterConn.xmlns) {
	    Engine.debug(Engine.DebugWarn,"XMLNS mismatch, required = " + Diameter.xmlns + ", received = " + xmlns);
	    return false;
	}
    }
    var children = xml.getChildren("c");
    if (!Array.isArray(children))
	return false;
    if (!children.length)
	return false;
    var capab_ops = "";
    var err = "";
    var appId = msg.appid;
    var state = this.state;
    for (var i = 0; i < children.length; i++) {
	var capab = children[i];
	if (capab) {
	    capab = capab.getText();
	    if (!capab) {
		if (children.length != 1) {
		    Engine.debug(Engine.DebugWarn,"Received request for closing with multiple <c/> elements, ignoring it");
		    return false;
		}
		state = "closing";
		break;
	    }
	    else {
		var capabObj = DiameterConn.capabs[capab];
		if (!capabObj) {
		    err = "Unsupported: " + capab;
		    state = "close";
		    break;
		}
	    }
	}
	if (appId) {
	    if (capabObj.appId != appId) {
		err = "Unsupported: " + capab;
		state = "close";
		break;
	    }
	}
	if (capabObj) {
	    if ("init" == this.state && "" == err)
		state = "accept";
	    if (capab_ops != "")
		capab_ops += ",";
	    capab_ops += capabObj.operations;
	    if (this.capabs != "")
		this.capabs += ", ";
	    this.capabs += capab;
	}
    }
    this.pendingOp = { state : state };
    if (msg.xml_decl.version)
	this.pendingOp.setDeclaration = true;

    if ("closing" == state) {
	msg.state = "closing";
	return true;
    }
    var xml = new XML("<m/>");

    switch (state) {
	case "accept":
	    xml.addChild("state","active");
	    xml.setAttribute("xmlns",DiameterConn.xmlns);
	    this.pendingOp.capab_ops = capab_ops;
	    break;
	case "close":
	    xml.addChild("state","inactive");
	    if (err != "")
		xml.addChild("error",err);
	    break;
	default:
	    break;
    }
    this.pendingOp.xml = xml;
    return true;
};

function checkDeclaration(msg)
{
    if (!msg)
	return false;
    var declVer = msg.xml_decl.version;
    if (declVer && "1.0" != declVer) {
	Engine.debug(Engine.DebugWarn,"Invalid version=" + declVer + " in XML declaration");
	return false;
    }
    var declEnc = msg.xml_decl.encoding;
    if (declEnc) {
	declEnc = declEnc.toLowerCase();
	if ("utf-8" != declEnc) {
	    Engine.debug(Engine.DebugWarn,"Invalid encoding=" + declEnc + " in XML declaration");
	    return false;
	}
    }
    return true;
}

function onDiamMngmt(msg)
{
    if ("diam_tcp" == msg.module)
	return false;
    var conn = DiameterConn.connList[msg.application];
    var xml = new XML(msg, "xml");
    switch (msg.operation) {
	case "create":
	    if (conn) {
		Engine.debug(Engine.DebugWarn,"Refusing attempt to create the same application=" + msg.application + " again");
		return false;
	    }
	    if (!checkDeclaration(msg))
		return false;
	    if (!xml) {
		Engine.debug(Engine.DebugWarn,"Refusing attempt to create an application, capabilities XML is missing");
		return false;
	    }
	    conn = new DiameterConn(msg.application);
	    if (!conn.handleXml(xml,msg,true)) {
		Engine.debug(Engine.DebugNote,"Failed to handle XML from management message");
		return false;
	    }
	    DiameterConn.connList[conn.name] = conn;
	    return true;
	case "close":
	    if (!conn) {
		Engine.debug(Engine.DebugInfo,"Closing request for unknown application=" + msg.application);
		return false;
	    }
	    delete DiameterConn.connList[msg.application];
	    return true;
	default:
	    if (!conn) {
		Engine.debug(Engine.DebugInfo,"Management request for unknown application=" + msg.application);
		return false;
	    }
	    if (!xml) {
		Engine.debug(Engine.DebugNote,"Failed to handle XML from management message");
		return false;
	    }
	    if (!conn.handleXml(xml,msg))
		return false;
	    return true;
    }
    return false;
}
// Helper that returns a left or right aligned fixed length string
function strFix(str,len,pad)
{
    if (null === str)
	str = "";
    if ("" == pad)
	pad = " ";
    if (len < 0) {
	// right aligned
	len = -len;
	if (str.length >= len)
	    return str.substr(str.length - len);
	while (str.length < len)
	    str = pad + str;
    }
    else {
	// left aligned
	if (str.length >= len)
	    return str.substr(0,len);
	while (str.length < len)
	    str += pad;
    }
    return str;
}

// Perform one command line completion
function oneCompletion(msg,str,part)
{
    if ("" != part && str.indexOf(part) != 0)
	return;;
    var ret = msg.retValue();
    if ("" != ret)
	ret += "\t";
    msg.retValue(ret + str);
}

function onComplete(msg,line,part)
{
    switch (line) {
	case undefined:
	case "":
	    oneCompletion(msg,"diam_tcp",part);
	    break;
	case "diam_tcp ":
	case "diam_tcp":
	    oneCompletion(msg,"list",part);
	    break;
    }
}

function onCommand(msg)
{
    if (!msg.line) {
	onComplete(msg,msg.partline,msg.partword);
	return false;
    }
    switch (msg.line) {
	case "diam_tcp list":
	    var tmp = "ID                     State      Capabilities\r\n";
	    tmp +=    "---------------------- ---------- ------------\r\n";
	    for (var conn of DiameterConn.connList) {
		tmp += strFix(conn.name,22) + " " + strFix(conn.state,11) + " " + conn.capabs + "\r\n";
	    }
	    msg.retValue(tmp);
	    return true;
    }
    return false;
}

function onInterval()
{
    for (var conn of DiameterConn.connList) {
	if (!conn.pendingOp)
	    continue;
	switch (conn.pendingOp.state) {
	    case "closing":
		var m = conn.buildMsg("diameter.query");
		m.query_params = "application";
		m.operation = "pending_reqs";
		if (!m.dispatch(true)) {
		    // failed to handle, close the connection
		    Engine.debug(Engine.DebugWarn,"Failed to query Diameter about status of requests for application="
			    + conn.name + ", marking it for closing");
		    conn.pendingOp.state = "close";
		    continue;
		}
		if (false == m.result) {
		    conn.pendingOp.state = "close";
		    conn.pendingOp.xml = new XML("<m><state>inactive</state></m>");
		}
		break;
	    case "close":
		var del = true;
		// intentional
	    case "accept":
		var m = conn.buildMsg("diameter.tcp",conn.pendingOp.xml);
		m.operation = conn.pendingOp.state;
		if (conn.pendingOp.capab_ops)
		    m.capab_ops = conn.pendingOp.capab_ops;
		if (conn.pendinOp.setDeclaration) {
		    m.xml_decl.version = DiameterConn.xmlDeclVersion;
		    m.xml_decl.encoding = DiameterConn.xmlDeclEncoding;
		}
		if (!m.dispatch(true)) {
		    Engine.debug(Engine.DebugWarn,"Failed to dispatch management message for application="
			    + conn.name + ", deleting application");
		    var del = true;
		}
		conn.state = conn.pendingOp.state;
		if (true == del) {
		     delete DiameterConn.connList[conn.name];
		     continue;
		}
		delete conn.pendingOp;
	}
    }
}

Engine.debugName("diam_tcp");
Message.trackName("diam_tcp");
Engine.debugEnabled(true);

Message.install(onDiamMngmt,"diameter.tcp",100);
Message.install(onCommand,"engine.command",100);

Engine.setInterval(onInterval, 1000)
