/**
 * enb_config.js
 * This file is part of the YATE Project http://YATE.null.ro
 *
 * Yet Another Telephony Engine - a fully featured software PBX and IVR
 * Copyright (C) 2014-2015 Null Team
 *
 * This software is distributed under multiple licenses;
 * see the COPYING file in the main directory for licensing
 * information for this specific distribution.
 *
 * This use of this software may be subject to additional restrictions.
 * See the LEGAL file in the main directory for details.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

//#pragma cache "true"
#pragma trace "cachegrind.out.enb_config"

#require "lib_str_util.js"
#require "sdr_config.js"

conf_node = "enb";
// define configuration files that can be set/get 
confs = ["gtp", "yateenb", "sdr", "ybladerf", "calibrate"];

// Class and reimplemented methods to configure gtp.conf 
GtpConfig = function()
{
    GenericConfig.apply(this);
    this.name = "gtp";
    this.file = "gtp";
    this.custom = "gtp-custom";
    this.error = new Object();
    this.skip_empty_params = new Object();
    this.sections = ["ran_u", "ran_c"];
    this.params_allowed_empty = ["addr4", "addr6"];
    this.params_required = {"ran_u":["type"]};
};
GtpConfig.prototype = new GenericConfig;
GtpConfig.prototype.validations = {
    "ran_u": {
	"addr4": {"callback": checkValidIP},
	"addr6": {"callback": checkValidIP},
	"type": {}  // gtp-access -- single value is possible and it will be set in prepareConfig
    }
};

GtpConfig.prototype.initConfig = function(params)
{
    if (debug)
	dumpObj("Entered GtpConfig.initConfig", params);

    var ran_u = params["ran_u"];
    if (!ran_u) {
	this.error.error = 401;
	this.error.reason = "Section 'ran_u' is required when setting 'gtp'.";
	return false;
    }

    if (!isFilled(ran_u.addr4) && !isFilled(ran_u.addr6)) {
	this.error.error = 401;
	this.error.reason = "Please set 'addr4' or 'addr6' when setting 'gtp'.";
	return false;
    }

    return true;
};

GtpConfig.prototype.genericPrepareConfig = GtpConfig.prototype.prepareConfig;

GtpConfig.prototype.prepareConfig = function(params)
{
    // add additional parameters to be written in ran_u section in gtp.conf
    // this are always fixed to this values, this is why they are not set from API
    params.ran_u.type = "gtp-access";
    params.ran_u.enabled = true;
    // unless set to false or to specific value, GTP will automatically listen on 0.0.0.0 / :: for ipv4/ipv6
    if (!params.ran_u.addr4)
	params.ran_u.addr4 = false;
    if (!params.ran_u.addr6)
	params.ran_u.addr6 = false;

    params.ran_c = {};
    params.ran_c.enabled = false;
    params.ran_c.type = "gtp-control";

    this.genericPrepareConfig(params);
};


// Class and reimplemented methods to configure yateenb.conf
EnbConfig = function()
{
    GenericConfig.apply(this);
    this.name = "enb";
    this.file = "yateenb";
    this.custom = "enb-custom";
    this.error = new Object();
    this.skip_empty_params = new Object();
    this.sections = ["general", "basic", "scheduler", "measurements", "bearers",  "radiohardware", "radio", "advanced", "uu-simulator", "uu-loopback", "test-enb"];
    this.params_allowed_empty = ["addr4", "addr6", "radio_driver", "address", "port", "pci", "earfcn", "broadcast", "max_pdu", "AddrSubnet", "AuxRoutingTable", "Name", "streams", "dscp", "antenna_type", "antenna_serial_number", "antenna_cable_type", "antenna_cable_length", "power_suply_type", "power_suply_serial_number", "location", "siteName", "antennaDirection", "mode"];
    this.params_required = {"basic": ["enodebId","MCC", "MNC", "TAC", "NID1", "NID2", "EARFCN", "Band","rootSequenceIndex", "prach-FreqOffset","groupAssignmentPUSCH"]};
};

EnbConfig.prototype = new GenericConfig;

#require "enb_fields.js"

// No need to save generic initConfig. It's just a stub 
EnbConfig.prototype.initConfig = function(params)
{
    if (debug)
	Engine.output("Entered EnbConfig.initConfig for object name " + this.name);

    if (!params.mme1) {
	this.error.reason = "Retrieving MME from DNS is not currently supported so setting at least one handconfigured MME is required. Missing required section 'mme'.";
	this.error.error = 402;
	return false;
    }

    //Srbx.mode == default, acknowledged, unacknowledged must check to not allow others Srbx.param to be set
    for (var i = 1; i < 3; i++) { 
	if (params.bearers["Srb" + i +".mode"] == "acknowledged") {
	    if (params.bearers["Srb" + i + ".rlcSnFieldLength"])  {
		this.error.reason = "The Srb" + i + ".mode acknowledged doesn't accept parameter: '" + "Srb" + i + ".rlcSnFieldLength" + "' in section 'bearers'.";
		this.error.error = 402;
		return false;
	    }
	}

	if (params.bearers["Srb" + i + ".mode"] == "default") {
	    if (i == 1) 
		var not_allowed = ["Srb1.rlcSnFieldLength", "Srb1.rlcTReordering", "Srb1.rlcTPollRetransmit",
		    "Srb1.rlcTStatusProhibit", "Srb1.rlcMaxRetxThreshold", "Srb1.rlcPollPdu", "Srb1.rlcPollByte"];
	    else
		var not_allowed = ["Srb2.rlcSnFieldLength", "Srb2.rlcTReordering", "Srb2.rlcTPollRetransmit",
		    "Srb2.rlcTStatusProhibit", "Srb2.rlcMaxRetxThreshold", "Srb2.rlcPollPdu", "Srb2.rlcPollByte"];
	}

	if (params.bearers["Srb" + i +".mode"] == "unacknowledged") {
	    if (i == 1)
		var not_allowed = ["Srb1.rlcMaxRetxThreshold", "Srb1.rlcPollPdu", "Srb1.rlcPollByte"];
	    else
		var not_allowed = ["Srb2.rlcMaxRetxThreshold", "Srb2.rlcPollPdu", "Srb2.rlcPollByte"];
	}

	for (var j = 0; j < not_allowed.length; j++) {
	    if (params.bearers[not_allowed[j]]) {
	        this.error.reason = "The Srb" + i + ".mode default doesn't accept parameter: '" + not_allowed[j] + "' in section 'bearers'.";
	        this.error.error = 401;
	        return false;
	    }
	}
    }
	
    return true;
};

EnbConfig.prototype.genericPrepareConfig = EnbConfig.prototype.prepareConfig;

EnbConfig.prototype.prepareConfig = function(params)
{
    if (debug)
	Engine.output("Entered EnbConfig.prepareConfig for object name " + this.name);

    var mme_validations = {
	"address": {"callback": checkValidIP},
	"local":   {"callback": checkValidIP},
	"streams": {"callback": checkValidInteger},
	"dscp":    {"select": ["expedited"]}
    };

    // see if we have any defined MMEs and modify sections and validations so this fields are written
    var mme_index = 0;
    var mme;
    var input = Array();
    while (true) {
	mme_index++;
	mme = "mme" + mme_index;
	if (!params[mme])
	    break;
	// add section name to this.sections and this.validations so section is written and verified
	if (debug)
	    Engine.output("Extended sections,validations and params_required in EnbConfig to include section '" + mme + "'");
	this.sections.push(mme);
	this.validations[mme] = mme_validations;
	this.params_required[mme] = ["address", "local"];
	if (mme_index>0)
	    input.push(params[mme]["address"]);
    }

    this.mme_dup = input;

    this.genericPrepareConfig();
};

EnbConfig.prototype.genericValidateConfig = EnbConfig.prototype.validateConfig;

EnbConfig.prototype.validateConfig = function(section_name,param_name,param_value,params)
{
    if (debug)
	Engine.output("Entered EnbConfig.validateConfig for object name " + this.name);

    if (!this.genericValidateConfig(section_name,param_name,param_value,params))
	return false;

    var duplicates = findMmeDuplicates(this.mme_dup);
    if (duplicates.length>0) {
	    this.error.reason = "Found duplicated MME Address(es): "+duplicates+" in section 'Core'.";
	    this.error.error = 402;
	    return false;
    }
   
    return true;
};

// Get yateenb.conf parameters
API.on_get_yateenb_node = function(params,msg)
{
    var enb = new EnbConfig;
    return API.on_get_generic_file(enb,msg);
};

// Configure yateenb.conf related parameters
API.on_set_yateenb_node = function(params,msg,setNode)
{
    var enb = new EnbConfig;
    var res = API.on_set_generic_file(enb,params,msg,setNode);
    if (res.reason)
	    return res;
    Engine.output("YateENB restart on node config: " + msg.received);
    Engine.restart();

    return res;
};

// Get gtp.conf params
API.on_get_gtp_node = function(params,msg)
{
    var gtp = new GtpConfig;
    return API.on_get_generic_file(gtp,msg);
};

// Configure gtp.conf params
API.on_set_gtp_node = function(params,msg,setNode)
{
    var gtp = new GtpConfig;
    return API.on_set_generic_file(gtp,params,msg,setNode);
};

// use generic get and set functions
API.on_set_enb_node = API.on_set_node;
API.on_get_enb_node = API.on_get_node;

// Retrieve available bands
API.on_get_bands = function(params,msg)
{
    var row;
    var band;
    var bnd;
    var fq_dl;
    var min_earf;
    var max_earf;
    var fq_ul;
    var earf_ofst;	
    var available_bands = [];

    var cpath = Engine.runParams("configpath");
    var bands_file = cpath + "/enb_bands.csv";

    if (!File.exists(bands_file)) {
	var err = "Missing " + bands_file + " file."; 
	return { error: 201, reason:err };
    }

    var bands = File.getContent(bands_file);
    bands = bands.split("\n");

    for (var i in bands) {
	row = bands[i];
	if (!row.length)
	    continue;
	row = row.split(",");
	if (row.length != 6) {
	    Engine.output("Found invalid row when retrieving bands: '" + bands[i] + "', length="+row.length);
	    continue;
	}
	bnd = row[0];       bnd = bnd.trim();              bnd = parseInt(bnd);
	fq_dl = row[1];     fq_dl = fq_dl.trim();          fq_dl = parseInt(fq_dl);
	min_earf = row[2];  min_earf = min_earf.trim();    min_earf = parseInt(min_earf);
	max_earf = row[3];  max_earf = max_earf.trim();    max_earf = parseInt(max_earf);
	fq_ul = row[4];     fq_ul = fq_ul.trim();          fq_ul = parseInt(fq_ul);
	earf_ofst = row[5]; earf_ofst = earf_ofst.trim();  earf_ofst = parseInt(earf_ofst);

	if (isNaN(bnd) || isNaN(fq_dl) || isNaN(min_earf) || isNaN(max_earf) || isNaN(fq_ul) || isNaN(earf_ofst) ) {
	    // in case the csv file has a header, exclude the first line
	    continue;
	}

	band = { band:bnd, frequency_dl:fq_dl, min_earfcn_dl:min_earf, max_earfcn_dl:max_earf, frequency_ul:fq_ul, earfcn_offset:earf_ofst };
	available_bands.push(band); 
    }

    return { name:"bands", object:available_bands };
};

// Handle rmanager commands
function onCommand(msg)
{
    if (!msg.line) {
	switch (msg.partline) {
	    case "debug":
		oneCompletion(msg,"enb_config",msg.partword);
		break;
	    case "debug enb_config":
		oneCompletion(msg,"on",msg.partword);
		oneCompletion(msg,"off",msg.partword);
		break;
	}
    }
    return false;
}

// Handle debugging command
function onDebug(msg)
{
    Engine.setDebug(msg.line);
    debug = Engine.debugEnabled();
    msg.retValue("enb_config debug " + Engine.debugEnabled() + " level " + Engine.debugLevel() + "\r\n");
    return true;
}

function findMmeDuplicates(arr)
{
    var i;
    var len = arr.length;
    var obj = {};
    var err = "";

    if (len < 1)
	return err;

    for (i = 0; i < len; i++) {
	if (!obj[arr[i]])
	    obj[arr[i]] = {};
	else
	    err += arr[i]+", ";
    }
    return err;
};

Engine.debugName("enb_config");
Message.trackName("enb_config");
loadCfg(true);

Message.install(onCommand,"engine.command",120);
Message.install(onDebug,"engine.debug",150,"module","enb_config");
Message.install(onReload,"engine.init",120);
Message.install(onApiRequest,"api.request",90,"type","config_enb");

/* vi: set ts=8 sw=4 sts=4 noet: */
